@@ -1,7 +1,7 @@
 /*
     gpt.[ch]
 
-    Copyright (C) 2000-2001 Dell Computer Corporation <Matt_Domsch@dell.com> 
+    Copyright (C) 2000-2001 Dell Computer Corporation <Matt_Domsch@dell.com>
 
     EFI GUID Partition Table handling
     Per Intel EFI Specification v1.02
@@ -18,8 +18,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #define _FILE_OFFSET_BITS 64
@@ -36,7 +35,9 @@
 #include <errno.h>
 #include <endian.h>
 #include <byteswap.h>
+#include <linux/fs.h>
 #include "crc32.h"
+#include "kpartx.h"
 
 #if BYTE_ORDER == LITTLE_ENDIAN
 #  define __le16_to_cpu(x) (x)
@@ -50,15 +51,23 @@
 #  define __cpu_to_le32(x) bswap_32(x)
 #endif
 
+#ifndef BLKGETLASTSECT
 #define BLKGETLASTSECT  _IO(0x12,108)   /* get last sector of block device */
+#endif
+#ifndef BLKGETSIZE
 #define BLKGETSIZE _IO(0x12,96)	        /* return device size */
+#endif
+#ifndef BLKSSZGET
 #define BLKSSZGET  _IO(0x12,104)	/* get block device sector size */
+#endif
+#ifndef BLKGETSIZE64
 #define BLKGETSIZE64 _IOR(0x12,114,sizeof(uint64_t))	/* return device size in bytes (u64 *arg) */
+#endif
 
 struct blkdev_ioctl_param {
-        unsigned int block;
-        size_t content_length;
-        char * block_contents;
+	unsigned int block;
+	size_t content_length;
+	char * block_contents;
 };
 
 /**
@@ -67,7 +76,7 @@ struct blkdev_ioctl_param {
  * @len - length of buf
  *
  * Description: Returns EFI-style CRC32 value for @buf
- * 
+ *
  * This function uses the little endian Ethernet polynomial
  * but seeds the function with ~0, and xor's with ~0 at the end.
  * Note, the EFI Specification, v1.02, has a reference to
@@ -97,7 +106,7 @@ is_pmbr_valid(legacy_mbr *mbr)
 	signature = (__le16_to_cpu(mbr->signature) == MSDOS_MBR_SIGNATURE);
 	for (i = 0; signature && i < 4; i++) {
 		if (mbr->partition[i].sys_type ==
-                    EFI_PMBR_OSTYPE_EFI_GPT) {
+		    EFI_PMBR_OSTYPE_EFI_GPT) {
 			found = 1;
 			break;
 		}
@@ -106,32 +115,13 @@ is_pmbr_valid(legacy_mbr *mbr)
 }
 
 
-/************************************************************
- * get_sector_size
- * Requires:
- *  - filedes is an open file descriptor, suitable for reading
- * Modifies: nothing
- * Returns:
- *  sector size, or 512.
- ************************************************************/
-static int
-get_sector_size(int filedes)
-{
-	int rc, sector_size = 512;
-
-	rc = ioctl(filedes, BLKSSZGET, &sector_size);
-	if (rc)
-		sector_size = 512;
-	return sector_size;
-}
-
 /************************************************************
  * _get_num_sectors
  * Requires:
  *  - filedes is an open file descriptor, suitable for reading
  * Modifies: nothing
  * Returns:
- *  Last LBA value on success 
+ *  Last LBA value on success
  *  0 on error
  *
  * Try getting BLKGETSIZE64 and BLKSSZGET first,
@@ -143,27 +133,21 @@ get_sector_size(int filedes)
 static uint64_t
 _get_num_sectors(int filedes)
 {
-	unsigned long sectors=0;
 	int rc;
-#if 0
-        uint64_t bytes=0;
+	uint64_t bytes=0;
 
- 	rc = ioctl(filedes, BLKGETSIZE64, &bytes);
+	rc = ioctl(filedes, BLKGETSIZE64, &bytes);
 	if (!rc)
 		return bytes / get_sector_size(filedes);
-#endif
-        rc = ioctl(filedes, BLKGETSIZE, &sectors);
-        if (rc)
-                return 0;
-        
-	return sectors;
+
+	return 0;
 }
 
 /************************************************************
  * last_lba(): return number of last logical block of device
- * 
+ *
  * @fd
- * 
+ *
  * Description: returns Last LBA value on success, 0 on error.
  * Notes: The value st_blocks gives the size of the file
  *        in 512-byte blocks, which is OK if
@@ -193,26 +177,26 @@ last_lba(int filedes)
 		sectors = 1;
 	}
 
-	return sectors - 1;
+	return sectors ? sectors - 1 : 0;
 }
 
 
 static ssize_t
 read_lastoddsector(int fd, uint64_t lba, void *buffer, size_t count)
 {
-        int rc;
-        struct blkdev_ioctl_param ioctl_param;
+	int rc;
+	struct blkdev_ioctl_param ioctl_param;
 
-        if (!buffer) return 0; 
+	if (!buffer) return 0;
 
-        ioctl_param.block = 0; /* read the last sector */
-        ioctl_param.content_length = count;
-        ioctl_param.block_contents = buffer;
+	ioctl_param.block = 0; /* read the last sector */
+	ioctl_param.content_length = count;
+	ioctl_param.block_contents = buffer;
 
-        rc = ioctl(fd, BLKGETLASTSECT, &ioctl_param);
-        if (rc == -1) perror("read failed");
+	rc = ioctl(fd, BLKGETLASTSECT, &ioctl_param);
+	if (rc == -1) perror("read failed");
 
-        return !rc;
+	return !rc;
 }
 
 static ssize_t
@@ -220,26 +204,32 @@ read_lba(int fd, uint64_t lba, void *buffer, size_t bytes)
 {
 	int sector_size = get_sector_size(fd);
 	off_t offset = lba * sector_size;
-        ssize_t bytesread;
+	uint64_t lastlba;
+	ssize_t bytesread;
 
-	lseek(fd, offset, SEEK_SET);
+	if (lseek(fd, offset, SEEK_SET) < 0)
+		return 0;
 	bytesread = read(fd, buffer, bytes);
 
-        /* Kludge.  This is necessary to read/write the last
-           block of an odd-sized disk, until Linux 2.5.x kernel fixes.
-           This is only used by gpt.c, and only to read
-           one sector, so we don't have to be fancy.
-        */
-        if (!bytesread && !(last_lba(fd) & 1) && lba == last_lba(fd)) {
-                bytesread = read_lastoddsector(fd, lba, buffer, bytes);
-        }
-        return bytesread;
+	lastlba = last_lba(fd);
+	if (!lastlba)
+		return bytesread;
+
+	/* Kludge.  This is necessary to read/write the last
+	   block of an odd-sized disk, until Linux 2.5.x kernel fixes.
+	   This is only used by gpt.c, and only to read
+	   one sector, so we don't have to be fancy.
+	*/
+	if (!bytesread && !(lastlba & 1) && lba == lastlba) {
+		bytesread = read_lastoddsector(fd, lba, buffer, bytes);
+	}
+	return bytesread;
 }
 
 /**
  * alloc_read_gpt_entries(): reads partition entries from disk
  * @fd  is an open file descriptor to the whole disk
- * @gpt is a buffer into which the GPT will be put  
+ * @gpt is a buffer into which the GPT will be put
  * Description: Returns ptes on success,  NULL on error.
  * Allocates space for PTEs based on information found in @gpt.
  * Notes: remember to free pte when you're done!
@@ -248,10 +238,10 @@ static gpt_entry *
 alloc_read_gpt_entries(int fd, gpt_header * gpt)
 {
 	gpt_entry *pte;
-        size_t count = __le32_to_cpu(gpt->num_partition_entries) *
-                __le32_to_cpu(gpt->sizeof_partition_entry);
+	size_t count = __le32_to_cpu(gpt->num_partition_entries) *
+		__le32_to_cpu(gpt->sizeof_partition_entry);
 
-        if (!count) return NULL;
+	if (!count) return NULL;
 
 	pte = (gpt_entry *)malloc(count);
 	if (!pte)
@@ -259,7 +249,7 @@ alloc_read_gpt_entries(int fd, gpt_header * gpt)
 	memset(pte, 0, count);
 
 	if (!read_lba(fd, __le64_to_cpu(gpt->partition_entry_lba), pte,
-                      count)) {
+		      count)) {
 		free(pte);
 		return NULL;
 	}
@@ -270,7 +260,7 @@ alloc_read_gpt_entries(int fd, gpt_header * gpt)
  * alloc_read_gpt_header(): Allocates GPT header, reads into it from disk
  * @fd  is an open file descriptor to the whole disk
  * @lba is the Logical Block Address of the partition table
- * 
+ *
  * Description: returns GPT header on success, NULL on error.   Allocates
  * and fills a GPT header starting at @ from @bdev.
  * Note: remember to free gpt when finished with it.
@@ -304,19 +294,19 @@ alloc_read_gpt_header(int fd, uint64_t lba)
  */
 static int
 is_gpt_valid(int fd, uint64_t lba,
-             gpt_header ** gpt, gpt_entry ** ptes)
+	     gpt_header ** gpt, gpt_entry ** ptes)
 {
 	int rc = 0;		/* default to not valid */
 	uint32_t crc, origcrc;
 
 	if (!gpt || !ptes)
-                return 0;
+		return 0;
 	if (!(*gpt = alloc_read_gpt_header(fd, lba)))
 		return 0;
 
 	/* Check the GUID Partition Table signature */
 	if (__le64_to_cpu((*gpt)->signature) != GPT_HEADER_SIGNATURE) {
-		/* 
+		/*
 		   printf("GUID Partition Table Header signature is wrong: %" PRIx64" != %" PRIx64 "\n",
 		   __le64_to_cpu((*gpt)->signature), GUID_PT_HEADER_SIGNATURE);
 		 */
@@ -343,13 +333,31 @@ is_gpt_valid(int fd, uint64_t lba,
 	if (__le64_to_cpu((*gpt)->my_lba) != lba) {
 		/*
 		printf( "my_lba % PRIx64 "x != lba %"PRIx64 "x.\n",
-		 		__le64_to_cpu((*gpt)->my_lba), lba);
+				__le64_to_cpu((*gpt)->my_lba), lba);
 		 */
 		free(*gpt);
 		*gpt = NULL;
 		return 0;
 	}
 
+	/* Check that sizeof_partition_entry has the correct value */
+	if (__le32_to_cpu((*gpt)->sizeof_partition_entry) != sizeof(gpt_entry)) {
+		// printf("GUID partition entry size check failed.\n");
+		free(*gpt);
+		*gpt = NULL;
+		return 0;
+	}
+
+
+	/* Check that sizeof_partition_entry has the correct value */
+	if (__le32_to_cpu((*gpt)->sizeof_partition_entry) != sizeof(gpt_entry)) {
+		// printf("GUID partition entry size check failed.\n");
+		free(*gpt);
+		*gpt = NULL;
+		return 0;
+	}
+
+
 	if (!(*ptes = alloc_read_gpt_entries(fd, *gpt))) {
 		free(*gpt);
 		*gpt = NULL;
@@ -358,7 +366,7 @@ is_gpt_valid(int fd, uint64_t lba,
 
 	/* Check the GUID Partition Entry Array CRC */
 	crc = efi_crc32(*ptes,
-                        __le32_to_cpu((*gpt)->num_partition_entries) *
+			__le32_to_cpu((*gpt)->num_partition_entries) *
 			__le32_to_cpu((*gpt)->sizeof_partition_entry));
 	if (crc != __le32_to_cpu((*gpt)->partition_entry_array_crc32)) {
 		// printf("GUID Partitition Entry Array CRC check failed.\n");
@@ -379,7 +387,7 @@ is_gpt_valid(int fd, uint64_t lba,
  * @lastlba is the last LBA number
  * Description: Returns nothing.  Sanity checks pgpt and agpt fields
  * and prints warnings on discrepancies.
- * 
+ *
  */
 static void
 compare_gpts(gpt_header *pgpt, gpt_header *agpt, uint64_t lastlba)
@@ -389,42 +397,42 @@ compare_gpts(gpt_header *pgpt, gpt_header *agpt, uint64_t lastlba)
 		return;
 	if (__le64_to_cpu(pgpt->my_lba) != __le64_to_cpu(agpt->alternate_lba)) {
 		error_found++;
-		fprintf(stderr, 
+		fprintf(stderr,
 		       "GPT:Primary header LBA != Alt. header alternate_lba\n");
 #ifdef DEBUG
 		fprintf(stderr,  "GPT:%" PRIx64 " != %" PRIx64 "\n",
 		       __le64_to_cpu(pgpt->my_lba),
-                       __le64_to_cpu(agpt->alternate_lba));
+		       __le64_to_cpu(agpt->alternate_lba));
 #endif
 	}
 	if (__le64_to_cpu(pgpt->alternate_lba) != __le64_to_cpu(agpt->my_lba)) {
 		error_found++;
-		fprintf(stderr, 
+		fprintf(stderr,
 		       "GPT:Primary header alternate_lba != Alt. header my_lba\n");
 #ifdef DEBUG
 		fprintf(stderr,  "GPT:%" PRIx64 " != %" PRIx64 "\n",
 		       __le64_to_cpu(pgpt->alternate_lba),
-                       __le64_to_cpu(agpt->my_lba));
+		       __le64_to_cpu(agpt->my_lba));
 #endif
 	}
 	if (__le64_to_cpu(pgpt->first_usable_lba) !=
-            __le64_to_cpu(agpt->first_usable_lba)) {
+	    __le64_to_cpu(agpt->first_usable_lba)) {
 		error_found++;
 		fprintf(stderr,  "GPT:first_usable_lbas don't match.\n");
 #ifdef DEBUG
 		fprintf(stderr,  "GPT:%" PRIx64 " != %" PRIx64 "\n",
 		       __le64_to_cpu(pgpt->first_usable_lba),
-                       __le64_to_cpu(agpt->first_usable_lba));
+		       __le64_to_cpu(agpt->first_usable_lba));
 #endif
 	}
 	if (__le64_to_cpu(pgpt->last_usable_lba) !=
-            __le64_to_cpu(agpt->last_usable_lba)) {
+	    __le64_to_cpu(agpt->last_usable_lba)) {
 		error_found++;
 		fprintf(stderr,  "GPT:last_usable_lbas don't match.\n");
 #ifdef DEBUG
 		fprintf(stderr,  "GPT:%" PRIx64 " != %" PRIx64 "\n",
 		       __le64_to_cpu(pgpt->last_usable_lba),
-                       __le64_to_cpu(agpt->last_usable_lba));
+		       __le64_to_cpu(agpt->last_usable_lba));
 #endif
 	}
 	if (efi_guidcmp(pgpt->disk_guid, agpt->disk_guid)) {
@@ -432,7 +440,7 @@ compare_gpts(gpt_header *pgpt, gpt_header *agpt, uint64_t lastlba)
 		fprintf(stderr,  "GPT:disk_guids don't match.\n");
 	}
 	if (__le32_to_cpu(pgpt->num_partition_entries) !=
-            __le32_to_cpu(agpt->num_partition_entries)) {
+	    __le32_to_cpu(agpt->num_partition_entries)) {
 		error_found++;
 		fprintf(stderr,  "GPT:num_partition_entries don't match: "
 		       "0x%x != 0x%x\n",
@@ -440,26 +448,26 @@ compare_gpts(gpt_header *pgpt, gpt_header *agpt, uint64_t lastlba)
 		       __le32_to_cpu(agpt->num_partition_entries));
 	}
 	if (__le32_to_cpu(pgpt->sizeof_partition_entry) !=
-            __le32_to_cpu(agpt->sizeof_partition_entry)) {
+	    __le32_to_cpu(agpt->sizeof_partition_entry)) {
 		error_found++;
-		fprintf(stderr, 
+		fprintf(stderr,
 		       "GPT:sizeof_partition_entry values don't match: "
 		       "0x%x != 0x%x\n",
-                       __le32_to_cpu(pgpt->sizeof_partition_entry),
+		       __le32_to_cpu(pgpt->sizeof_partition_entry),
 		       __le32_to_cpu(agpt->sizeof_partition_entry));
 	}
 	if (__le32_to_cpu(pgpt->partition_entry_array_crc32) !=
-            __le32_to_cpu(agpt->partition_entry_array_crc32)) {
+	    __le32_to_cpu(agpt->partition_entry_array_crc32)) {
 		error_found++;
-		fprintf(stderr, 
+		fprintf(stderr,
 		       "GPT:partition_entry_array_crc32 values don't match: "
 		       "0x%x != 0x%x\n",
-                       __le32_to_cpu(pgpt->partition_entry_array_crc32),
+		       __le32_to_cpu(pgpt->partition_entry_array_crc32),
 		       __le32_to_cpu(agpt->partition_entry_array_crc32));
 	}
 	if (__le64_to_cpu(pgpt->alternate_lba) != lastlba) {
 		error_found++;
-		fprintf(stderr, 
+		fprintf(stderr,
 		       "GPT:Primary header thinks Alt. header is not at the end of the disk.\n");
 #ifdef DEBUG
 		fprintf(stderr,  "GPT:%" PRIx64 " != %" PRIx64 "\n",
@@ -469,7 +477,7 @@ compare_gpts(gpt_header *pgpt, gpt_header *agpt, uint64_t lastlba)
 
 	if (__le64_to_cpu(agpt->my_lba) != lastlba) {
 		error_found++;
-		fprintf(stderr, 
+		fprintf(stderr,
 		       "GPT:Alternate GPT header not at the end of the disk.\n");
 #ifdef DEBUG
 		fprintf(stderr,  "GPT:%" PRIx64 " != %" PRIx64 "\n",
@@ -478,7 +486,7 @@ compare_gpts(gpt_header *pgpt, gpt_header *agpt, uint64_t lastlba)
 	}
 
 	if (error_found)
-		fprintf(stderr, 
+		fprintf(stderr,
 		       "GPT: Use GNU Parted to correct GPT errors.\n");
 	return;
 }
@@ -496,7 +504,6 @@ compare_gpts(gpt_header *pgpt, gpt_header *agpt, uint64_t lastlba)
 static int
 find_valid_gpt(int fd, gpt_header ** gpt, gpt_entry ** ptes)
 {
-        extern int force_gpt;
 	int good_pgpt = 0, good_agpt = 0, good_pmbr = 0;
 	gpt_header *pgpt = NULL, *agpt = NULL;
 	gpt_entry *pptes = NULL, *aptes = NULL;
@@ -505,96 +512,97 @@ find_valid_gpt(int fd, gpt_header ** gpt, gpt_entry ** ptes)
 	if (!gpt || !ptes)
 		return 0;
 
-	lastlba = last_lba(fd);
+	if (!(lastlba = last_lba(fd)))
+		return 0;
 	good_pgpt = is_gpt_valid(fd, GPT_PRIMARY_PARTITION_TABLE_LBA,
 				 &pgpt, &pptes);
-        if (good_pgpt) {
+	if (good_pgpt) {
 		good_agpt = is_gpt_valid(fd,
-                                         __le64_to_cpu(pgpt->alternate_lba),
+					 __le64_to_cpu(pgpt->alternate_lba),
 					 &agpt, &aptes);
-                if (!good_agpt) {
-                        good_agpt = is_gpt_valid(fd, lastlba,
-                                                 &agpt, &aptes);
-                }
-        }
-        else {
-                good_agpt = is_gpt_valid(fd, lastlba,
-                                         &agpt, &aptes);
-        }
+		if (!good_agpt) {
+			good_agpt = is_gpt_valid(fd, lastlba,
+						 &agpt, &aptes);
+		}
+	}
+	else {
+		good_agpt = is_gpt_valid(fd, lastlba,
+					 &agpt, &aptes);
+	}
 
-        /* The obviously unsuccessful case */
-        if (!good_pgpt && !good_agpt) {
-                goto fail;
-        }
+	/* The obviously unsuccessful case */
+	if (!good_pgpt && !good_agpt) {
+		goto fail;
+	}
 
 	/* This will be added to the EFI Spec. per Intel after v1.02. */
-        legacymbr = malloc(sizeof (*legacymbr));
-        if (legacymbr) {
-                memset(legacymbr, 0, sizeof (*legacymbr));
-                read_lba(fd, 0, (uint8_t *) legacymbr,
-                         sizeof (*legacymbr));
-                good_pmbr = is_pmbr_valid(legacymbr);
-                free(legacymbr);
-                legacymbr=NULL;
-        }
+	legacymbr = malloc(sizeof (*legacymbr));
+	if (legacymbr) {
+		memset(legacymbr, 0, sizeof (*legacymbr));
+		read_lba(fd, 0, (uint8_t *) legacymbr,
+			 sizeof (*legacymbr));
+		good_pmbr = is_pmbr_valid(legacymbr);
+		free(legacymbr);
+		legacymbr=NULL;
+	}
 
-        /* Failure due to bad PMBR */
-        if ((good_pgpt || good_agpt) && !good_pmbr && !force_gpt) {
-                fprintf(stderr,
-                       "  Warning: Disk has a valid GPT signature "
-                       "but invalid PMBR.\n"
-                       "  Assuming this disk is *not* a GPT disk anymore.\n"
-                       "  Use gpt kernel option to override.  "
-                       "Use GNU Parted to correct disk.\n");
-                goto fail;
-        }
+	/* Failure due to bad PMBR */
+	if ((good_pgpt || good_agpt) && !good_pmbr && !force_gpt) {
+		fprintf(stderr,
+		       "  Warning: Disk has a valid GPT signature "
+		       "but invalid PMBR.\n"
+		       "  Assuming this disk is *not* a GPT disk anymore.\n"
+		       "  Use gpt kernel option to override.  "
+		       "Use GNU Parted to correct disk.\n");
+		goto fail;
+	}
 
-        /* Would fail due to bad PMBR, but force GPT anyhow */
-        if ((good_pgpt || good_agpt) && !good_pmbr && force_gpt) {
-                fprintf(stderr, 
-                       "  Warning: Disk has a valid GPT signature but "
-                       "invalid PMBR.\n"
-                       "  Use GNU Parted to correct disk.\n"
-                       "  gpt option taken, disk treated as GPT.\n");
-        }
+	/* Would fail due to bad PMBR, but force GPT anyhow */
+	if ((good_pgpt || good_agpt) && !good_pmbr && force_gpt) {
+		fprintf(stderr,
+		       "  Warning: Disk has a valid GPT signature but "
+		       "invalid PMBR.\n"
+		       "  Use GNU Parted to correct disk.\n"
+		       "  gpt option taken, disk treated as GPT.\n");
+	}
 
-        compare_gpts(pgpt, agpt, lastlba);
+	compare_gpts(pgpt, agpt, lastlba);
 
-        /* The good cases */
-        if (good_pgpt && (good_pmbr || force_gpt)) {
-                *gpt  = pgpt;
-                *ptes = pptes;
-                if (agpt)  { free(agpt);   agpt = NULL; }
-                if (aptes) { free(aptes); aptes = NULL; }
-                if (!good_agpt) {
-                        fprintf(stderr, 
+	/* The good cases */
+	if (good_pgpt && (good_pmbr || force_gpt)) {
+		*gpt  = pgpt;
+		*ptes = pptes;
+		if (agpt)  { free(agpt);   agpt = NULL; }
+		if (aptes) { free(aptes); aptes = NULL; }
+		if (!good_agpt) {
+			fprintf(stderr,
 			       "Alternate GPT is invalid, "
-                               "using primary GPT.\n");
-                }
-                return 1;
-        }
-        else if (good_agpt && (good_pmbr || force_gpt)) {
-                *gpt  = agpt;
-                *ptes = aptes;
-                if (pgpt)  { free(pgpt);   pgpt = NULL; }
-                if (pptes) { free(pptes); pptes = NULL; }
-                fprintf(stderr, 
-                       "Primary GPT is invalid, using alternate GPT.\n");
-                return 1;
-        }
+			       "using primary GPT.\n");
+		}
+		return 1;
+	}
+	else if (good_agpt && (good_pmbr || force_gpt)) {
+		*gpt  = agpt;
+		*ptes = aptes;
+		if (pgpt)  { free(pgpt);   pgpt = NULL; }
+		if (pptes) { free(pptes); pptes = NULL; }
+		fprintf(stderr,
+		       "Primary GPT is invalid, using alternate GPT.\n");
+		return 1;
+	}
 
- fail:
-        if (pgpt)  { free(pgpt);   pgpt=NULL; }
-        if (agpt)  { free(agpt);   agpt=NULL; }
-        if (pptes) { free(pptes); pptes=NULL; }
-        if (aptes) { free(aptes); aptes=NULL; }
-        *gpt = NULL;
-        *ptes = NULL;
-        return 0;
+fail:
+	if (pgpt)  { free(pgpt);   pgpt=NULL; }
+	if (agpt)  { free(agpt);   agpt=NULL; }
+	if (pptes) { free(pptes); pptes=NULL; }
+	if (aptes) { free(aptes); aptes=NULL; }
+	*gpt = NULL;
+	*ptes = NULL;
+	return 0;
 }
 
 /**
- * read_gpt_pt() 
+ * read_gpt_pt()
  * @fd
  * @all - slice with start/size of whole disk
  *
@@ -609,7 +617,8 @@ read_gpt_pt (int fd, struct slice all, struct slice *sp, int ns)
 	gpt_entry *ptes = NULL;
 	uint32_t i;
 	int n = 0;
-        int last_used_index=-1;
+	int last_used_index=-1;
+	int sector_size_mul = get_sector_size(fd)/512;
 
 	if (!find_valid_gpt (fd, &gpt, &ptes) || !gpt || !ptes) {
 		if (gpt)
@@ -625,10 +634,12 @@ read_gpt_pt (int fd, struct slice all, struct slice *sp, int ns)
 			sp[n].size = 0;
 			n++;
 		} else {
-			sp[n].start = __le64_to_cpu(ptes[i].starting_lba);
-			sp[n].size  = __le64_to_cpu(ptes[i].ending_lba) -
-				__le64_to_cpu(ptes[i].starting_lba) + 1;
-                        last_used_index=n;
+			sp[n].start = sector_size_mul *
+				      __le64_to_cpu(ptes[i].starting_lba);
+			sp[n].size  = sector_size_mul *
+				      (__le64_to_cpu(ptes[i].ending_lba) -
+				       __le64_to_cpu(ptes[i].starting_lba) + 1);
+			last_used_index=n;
 			n++;
 		}
 	}
