@@ -2,7 +2,7 @@
 /* Added vfs mount options - aeb - 960223 */
 /* Removed lomount - aeb - 960224 */
 
-/* 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
+/* 1999-02-22 Arkadiusz Miśkiewicz <misiek@pld.ORG.PL>
  * - added Native Language Support
  * Sun Mar 21 1999 - Arnaldo Carvalho de Melo <acme@conectiva.com.br>
  * - fixed strerr(errno) in gettext calls
@@ -25,8 +25,9 @@
 #include <sys/ioctl.h>
 #include <sys/stat.h>
 #include <sys/mman.h>
-#include <sysmacros.h>
-#include <asm/posix_types.h>
+#include <sys/types.h>
+#include <dirent.h>
+#include "sysmacros.h"
 #include <linux/loop.h>
 
 #include "lopart.h"
@@ -37,7 +38,7 @@
 #endif
 
 #if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__) \
-        && !defined (__s390x__)
+	&& !defined (__s390x__)
 #define int2ptr(x)	((void *) ((int) x))
 #else
 #define int2ptr(x)	((void *) ((long) x))
@@ -75,9 +76,9 @@ is_loop_device (const char *device)
 	loopmajor = 0;
 
 	if ((procdev = fopen(PROC_DEVICES, "r")) != NULL) {
-		
+
 		while (fgets (line, sizeof(line), procdev)) {
-			
+
 			if ((cp = strstr (line, " loop\n")) != NULL) {
 				*cp='\0';
 				loopmajor=atoi(line);
@@ -98,94 +99,96 @@ is_loop_device (const char *device)
 extern char *
 find_loop_by_file (const char * filename)
 {
-	char dev[64];
-	char *loop_formats[] = { "/dev/loop%d", "/dev/loop/%d" };
-	int i, j, fd;
+	DIR *dir;
+	struct dirent *dent;
+	char dev[64], *found = NULL;
+	int fd;
 	struct stat statbuf;
 	struct loop_info loopinfo;
 
-	for (j = 0; j < SIZE(loop_formats); j++) {
+	dir = opendir("/dev");
+	if (!dir)
+		return NULL;
 
-		for (i = 0; i < 256; i++) {
-			sprintf (dev, loop_formats[j], i);
+	while ((dent = readdir(dir)) != NULL) {
+		if (strncmp(dent->d_name,"loop",4))
+			continue;
+		if (!strcmp(dent->d_name, "loop-control"))
+			continue;
+		sprintf(dev, "/dev/%s", dent->d_name);
 
-			if (stat (dev, &statbuf) != 0 ||
-			    !S_ISBLK(statbuf.st_mode))
-				continue;
-
-			fd = open (dev, O_RDONLY);
-
-			if (fd < 0)
-				break;
-
-			if (ioctl (fd, LOOP_GET_STATUS, &loopinfo) != 0) {
-				close (fd);
-				continue;
-			}
-
-			if (0 == strcmp(filename, loopinfo.lo_name)) {
-				close (fd);
-				return xstrdup(dev); /*found */
-			}
+		fd = open (dev, O_RDONLY);
+		if (fd < 0)
+			break;
 
+		if (fstat (fd, &statbuf) != 0 ||
+		    !S_ISBLK(statbuf.st_mode)) {
 			close (fd);
 			continue;
 		}
+
+		if (ioctl (fd, LOOP_GET_STATUS, &loopinfo) != 0) {
+			close (fd);
+			continue;
+		}
+
+		if (0 == strcmp(filename, loopinfo.lo_name)) {
+			close (fd);
+			found = xstrdup(dev);
+			break;
+		}
+
+		close (fd);
 	}
-	return NULL;
+	closedir(dir);
+	return found;
 }
 
 extern char *
 find_unused_loop_device (void)
 {
-	/* Just creating a device, say in /tmp, is probably a bad idea -
-	   people might have problems with backup or so.
-	   So, we just try /dev/loop[0-7]. */
-
-	char dev[20];
-	char *loop_formats[] = { "/dev/loop%d", "/dev/loop/%d" };
-	int i, j, fd, first = 0, somedev = 0, someloop = 0, loop_known = 0;
+	char dev[20], *next_loop_dev = NULL;
+	int fd, next_loop = 0, somedev = 0, someloop = 0, loop_known = 0;
 	struct stat statbuf;
 	struct loop_info loopinfo;
 	FILE *procdev;
 
-	if (stat("/dev/loop-control", &statbuf) == 0 &&
-	    S_ISCHR(statbuf.st_mode)) {
-		fd = open("/dev/loop-control", O_RDWR);
-		if (fd >= 0)
-			first = ioctl(fd, LOOP_CTL_GET_FREE);
-		close(fd);
-		if (first < 0)
-			first = 0;
-	}
-	for (j = 0; j < SIZE(loop_formats); j++) {
+	while (next_loop_dev == NULL) {
+		if (stat("/dev/loop-control", &statbuf) == 0 &&
+		    S_ISCHR(statbuf.st_mode)) {
+			int next_loop_fd;
 
-	    for(i = first; i < 256; i++) {
-		sprintf(dev, loop_formats[j], i);
+			next_loop_fd = open("/dev/loop-control", O_RDWR);
+			if (next_loop_fd < 0)
+				return NULL;
+			next_loop = ioctl(next_loop_fd, LOOP_CTL_GET_FREE);
+			close(next_loop_fd);
+			if (next_loop < 0)
+				return NULL;
+		}
 
-		if (stat (dev, &statbuf) == 0 && S_ISBLK(statbuf.st_mode)) {
-			somedev++;
-			fd = open (dev, O_RDONLY);
-
-			if (fd >= 0) {
+		sprintf(dev, "/dev/loop%d", next_loop);
 
+		fd = open (dev, O_RDONLY);
+		if (fd >= 0) {
+			if (fstat (fd, &statbuf) == 0 &&
+			    S_ISBLK(statbuf.st_mode)) {
+				somedev++;
 				if(ioctl (fd, LOOP_GET_STATUS, &loopinfo) == 0)
 					someloop++;		/* in use */
+				else if (errno == ENXIO)
+					next_loop_dev = xstrdup(dev);
 
-				else if (errno == ENXIO) {
-					close (fd);
-					return xstrdup(dev);/* probably free */
-				}
-
-				close (fd);
 			}
-			
+			close (fd);
+
 			/* continue trying as long as devices exist */
 			continue;
 		}
 		break;
-	    }
 	}
+	if (next_loop_dev)
+		return next_loop_dev;
 
 	/* Nothing found. Why not? */
 	if ((procdev = fopen(PROC_DEVICES, "r")) != NULL) {
@@ -208,29 +211,24 @@ find_unused_loop_device (void)
 		fprintf(stderr, "mount: could not find any device /dev/loop#");
 
 	else if (!someloop) {
-
-	    if (loop_known == 1)
-		fprintf(stderr,
-		    "mount: Could not find any loop device.\n"
-		    "       Maybe /dev/loop# has a wrong major number?");
-	    
-	    else if (loop_known == -1)
-		fprintf(stderr,
-		    "mount: Could not find any loop device, and, according to %s,\n"
-		    "       this kernel does not know about the loop device.\n"
-		    "       (If so, then recompile or `modprobe loop'.)",
-		      PROC_DEVICES);
-
-	    else
-		fprintf(stderr,
-		    "mount: Could not find any loop device. Maybe this kernel does not know\n"
-		    "       about the loop device (then recompile or `modprobe loop'), or\n"
-		    "       maybe /dev/loop# has the wrong major number?");
-
+		if (loop_known == 1)
+			fprintf(stderr,
+				"mount: Could not find any loop device.\n"
+				"       Maybe /dev/loop# has a wrong major number?");
+		else if (loop_known == -1)
+			fprintf(stderr,
+				"mount: Could not find any loop device, and, according to %s,\n"
+				"       this kernel does not know about the loop device.\n"
+				"       (If so, then recompile or `modprobe loop'.)",
+				PROC_DEVICES);
+		else
+			fprintf(stderr,
+				"mount: Could not find any loop device. Maybe this kernel does not know\n"
+				"       about the loop device (then recompile or `modprobe loop'), or\n"
+				"       maybe /dev/loop# has the wrong major number?");
 	} else
 		fprintf(stderr, "mount: could not find any free loop device");
-	
-	return 0;
+	return NULL;
 }
 
 extern int
@@ -253,6 +251,7 @@ set_loop (const char *device, const char *file, int offset, int *loopro)
 	}
 
 	if ((fd = open (device, mode)) < 0) {
+		close(ffd);
 		perror (device);
 		return 1;
 	}
@@ -285,10 +284,10 @@ set_loop (const char *device, const char *file, int offset, int *loopro)
 	return 0;
 }
 
-extern int 
+extern int
 del_loop (const char *device)
 {
-	int retries = 3;
+	int retries = 5;
 	int fd;
 
 	if ((fd = open (device, O_RDONLY)) < 0) {
