@@ -80,6 +80,8 @@ set_ble_device (vector blist, char * vendor, char * product, int origin)
 		if (regcomp(&ble->vendor_reg, vendor,
 			    REG_EXTENDED|REG_NOSUB)) {
 			FREE(vendor);
+			if (product)
+				FREE(product);
 			return 1;
 		}
 		ble->vendor = vendor;
@@ -88,6 +90,10 @@ set_ble_device (vector blist, char * vendor, char * product, int origin)
 		if (regcomp(&ble->product_reg, product,
 			    REG_EXTENDED|REG_NOSUB)) {
 			FREE(product);
+			if (vendor) {
+				ble->vendor = NULL;
+				FREE(vendor);
+			}
 			return 1;
 		}
 		ble->product = product;
@@ -129,8 +135,12 @@ _blacklist_exceptions_device(vector elist, char * vendor, char * product)
 	struct blentry_device * ble;
 
 	vector_foreach_slot (elist, ble, i) {
-		if (!regexec(&ble->vendor_reg, vendor, 0, NULL, 0) &&
-		    !regexec(&ble->product_reg, product, 0, NULL, 0))
+		if (!ble->vendor && !ble->product)
+			continue;
+		if ((!ble->vendor ||
+		     !regexec(&ble->vendor_reg, vendor, 0, NULL, 0)) &&
+		    (!ble->product ||
+		     !regexec(&ble->product_reg, product, 0, NULL, 0)))
 			return 1;
 	}
 	return 0;
@@ -143,8 +153,12 @@ _blacklist_device (vector blist, char * vendor, char * product)
 	struct blentry_device * ble;
 
 	vector_foreach_slot (blist, ble, i) {
-		if (!regexec(&ble->vendor_reg, vendor, 0, NULL, 0) &&
-		    !regexec(&ble->product_reg, product, 0, NULL, 0))
+		if (!ble->vendor && !ble->product)
+			continue;
+		if ((!ble->vendor ||
+		     !regexec(&ble->vendor_reg, vendor, 0, NULL, 0)) &&
+		    (!ble->product ||
+		     !regexec(&ble->product_reg, product, 0, NULL, 0)))
 			return 1;
 	}
 	return 0;
@@ -158,25 +172,25 @@ setup_default_blist (struct config * conf)
 	char * str;
 	int i;
 
-	str = STRDUP("^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*");
+	str = STRDUP("^(ram|raw|loop|fd|md|dm-|sr|scd|st|dcssblk)[0-9]");
 	if (!str)
 		return 1;
 	if (store_ble(conf->blist_devnode, str, ORIGIN_DEFAULT))
 		return 1;
 
-	str = STRDUP("^(td|hd)[a-z]");
+	str = STRDUP("^(td|hd|vd)[a-z]");
 	if (!str)
 		return 1;
 	if (store_ble(conf->blist_devnode, str, ORIGIN_DEFAULT))
 		return 1;
 
-	str = STRDUP("^dcssblk[0-9]*");
+	str = STRDUP("^nvme");
 	if (!str)
 		return 1;
 	if (store_ble(conf->blist_devnode, str, ORIGIN_DEFAULT))
 		return 1;
 
-	str = STRDUP("(ID_SCSI_VPD|ID_WWN)");
+	str = STRDUP("(SCSI_IDENT_|ID_WWN)");
 	if (!str)
 		return 1;
 	if (store_ble(conf->elist_property, str, ORIGIN_DEFAULT))
@@ -190,12 +204,13 @@ setup_default_blist (struct config * conf)
 			if (alloc_ble_device(conf->blist_device))
 				return 1;
 			ble = VECTOR_SLOT(conf->blist_device,
-					  VECTOR_SIZE(conf->blist_device) -1);
+					  VECTOR_SIZE(conf->blist_device) - 1);
 			if (set_ble_device(conf->blist_device,
 					   STRDUP(hwe->vendor),
 					   STRDUP(hwe->bl_product),
 					   ORIGIN_DEFAULT)) {
 				FREE(ble);
+				vector_del_slot(conf->blist_device, VECTOR_SIZE(conf->blist_device) - 1);
 				return 1;
 			}
 		}
@@ -207,6 +222,8 @@ setup_default_blist (struct config * conf)
 	if (vendor && product)						\
 		condlog(3, "%s: (%s:%s) %s %s",				\
 			dev, vendor, product, (M), (S));		\
+	else if (wwid && !dev)						\
+		condlog(3, "%s: %s %s", wwid, (M), (S));		\
 	else if (wwid)							\
 		condlog(3, "%s: %s %s %s", dev, (M), wwid, (S));	\
 	else if (env)							\
@@ -307,10 +324,10 @@ _filter_wwid (vector blist, vector elist, char * wwid)
 }
 
 int
-filter_wwid (vector blist, vector elist, char * wwid)
+filter_wwid (vector blist, vector elist, char * wwid, char * dev)
 {
 	int r = _filter_wwid(blist, elist, wwid);
-	log_filter(NULL, NULL, NULL, wwid, NULL, r);
+	log_filter(dev, NULL, NULL, wwid, NULL, r);
 	return r;
 }
 
