@@ -6,6 +6,9 @@
 #include <stdio.h>
 #include <string.h>
 #include <libudev.h>
+#include <dirent.h>
+#include <limits.h>
+#include <errno.h>
 
 #include "checkers.h"
 #include "memory.h"
@@ -21,6 +24,7 @@
 #include "defaults.h"
 #include "prio.h"
 #include "devmapper.h"
+#include "mpath_cmd.h"
 
 static int
 hwe_strmatch (struct hwentry *hwe1, struct hwentry *hwe2)
@@ -76,7 +80,8 @@ hwe_regmatch (struct hwentry *hwe1, struct hwentry *hwe2)
 	    regcomp(&rre, hwe1->revision, REG_EXTENDED|REG_NOSUB))
 		goto out_pre;
 
-	if ((!hwe1->vendor || !hwe2->vendor ||
+	if ((hwe2->vendor || hwe2->product || hwe2->revision) &&
+	    (!hwe1->vendor || !hwe2->vendor ||
 	     !regexec(&vre, hwe2->vendor, 0, NULL, 0)) &&
 	    (!hwe1->product || !hwe2->product ||
 	     !regexec(&pre, hwe2->product, 0, NULL, 0)) &&
@@ -121,7 +126,7 @@ find_hwe (vector hwtable, char * vendor, char * product, char * revision)
 }
 
 extern struct mpentry *
-find_mpe (char * wwid)
+find_mpe (vector mptable, char * wwid)
 {
 	int i;
 	struct mpentry * mpe;
@@ -129,7 +134,7 @@ find_mpe (char * wwid)
 	if (!wwid)
 		return NULL;
 
-	vector_foreach_slot (conf->mptable, mpe, i)
+	vector_foreach_slot (mptable, mpe, i)
 		if (mpe->wwid && !strcmp(mpe->wwid, wwid))
 			return mpe;
 
@@ -137,7 +142,7 @@ find_mpe (char * wwid)
 }
 
 extern char *
-get_mpe_wwid (char * alias)
+get_mpe_wwid (vector mptable, char * alias)
 {
 	int i;
 	struct mpentry * mpe;
@@ -145,7 +150,7 @@ get_mpe_wwid (char * alias)
 	if (!alias)
 		return NULL;
 
-	vector_foreach_slot (conf->mptable, mpe, i)
+	vector_foreach_slot (mptable, mpe, i)
 		if (mpe->alias && strcmp(mpe->alias, alias) == 0)
 			return mpe->wwid;
 
@@ -340,6 +345,10 @@ merge_hwe (struct hwentry * dst, struct hwentry * src)
 	merge_num(user_friendly_names);
 	merge_num(retain_hwhandler);
 	merge_num(detect_prio);
+	merge_num(deferred_remove);
+	merge_num(delay_watch_checks);
+	merge_num(delay_wait_checks);
+	merge_num(skip_kpartx);
 
 	/*
 	 * Make sure features is consistent with
@@ -471,9 +480,6 @@ free_config (struct config * conf)
 	if (!conf)
 		return;
 
-	if (conf->dev)
-		FREE(conf->dev);
-
 	if (conf->multipath_dir)
 		FREE(conf->multipath_dir);
 
@@ -497,17 +503,24 @@ free_config (struct config * conf)
 
 	if (conf->wwids_file)
 		FREE(conf->wwids_file);
+
 	if (conf->prio_name)
 		FREE(conf->prio_name);
 
 	if (conf->alias_prefix)
 		FREE(conf->alias_prefix);
+	if (conf->partition_delim)
+		FREE(conf->partition_delim);
 
 	if (conf->prio_args)
 		FREE(conf->prio_args);
 
 	if (conf->checker_name)
 		FREE(conf->checker_name);
+
+	if (conf->config_dir)
+		FREE(conf->config_dir);
+
 	if (conf->reservation_key)
 		FREE(conf->reservation_key);
 
@@ -523,18 +536,55 @@ free_config (struct config * conf)
 
 	free_mptable(conf->mptable);
 	free_hwtable(conf->hwtable);
+	free_hwe(conf->overrides);
 	free_keywords(conf->keywords);
 	FREE(conf);
 }
 
-int
-load_config (char * file, struct udev *udev)
+/* if multipath fails to process the config directory, it should continue,
+ * with just a warning message */
+static void
+process_config_dir(struct config *conf, vector keywords, char *dir)
 {
-	if (!conf)
-		conf = alloc_config();
+	struct dirent **namelist;
+	int i, n;
+	char path[LINE_MAX];
+	int old_hwtable_size;
 
-	if (!conf || !udev)
-		return 1;
+	if (dir[0] != '/') {
+		condlog(1, "config_dir '%s' must be a fully qualified path",
+			dir);
+		return;
+	}
+	n = scandir(dir, &namelist, NULL, alphasort);
+	if (n < 0) {
+		if (errno == ENOENT)
+			condlog(3, "No configuration dir '%s'", dir);
+		else
+			condlog(0, "couldn't open configuration dir '%s': %s",
+				dir, strerror(errno));
+		return;
+	}
+	for (i = 0; i < n; i++) {
+		if (!strstr(namelist[i]->d_name, ".conf"))
+			continue;
+		old_hwtable_size = VECTOR_SIZE(conf->hwtable);
+		snprintf(path, LINE_MAX, "%s/%s", dir, namelist[i]->d_name);
+		path[LINE_MAX-1] = '\0';
+		process_file(conf, path);
+		if (VECTOR_SIZE(conf->hwtable) > old_hwtable_size)
+			factorize_hwtable(conf->hwtable, old_hwtable_size);
+
+	}
+}
+
+struct config *
+load_config (char * file)
+{
+	struct config *conf = alloc_config();
+
+	if (!conf)
+		return NULL;
 
 	/*
 	 * internal defaults
@@ -542,25 +592,34 @@ load_config (char * file, struct udev *udev)
 	if (!conf->verbosity)
 		conf->verbosity = DEFAULT_VERBOSITY;
 
-	conf->udev = udev;
-	conf->dev_type = DEV_NONE;
 	conf->minio = DEFAULT_MINIO;
 	conf->minio_rq = DEFAULT_MINIO_RQ;
 	get_sys_max_fds(&conf->max_fds);
 	conf->bindings_file = set_default(DEFAULT_BINDINGS_FILE);
 	conf->wwids_file = set_default(DEFAULT_WWIDS_FILE);
-	conf->bindings_read_only = 0;
 	conf->multipath_dir = set_default(DEFAULT_MULTIPATHDIR);
 	conf->features = set_default(DEFAULT_FEATURES);
-	conf->flush_on_last_del = 0;
+	conf->flush_on_last_del = DEFAULT_FLUSH;
 	conf->attribute_flags = 0;
 	conf->reassign_maps = DEFAULT_REASSIGN_MAPS;
 	conf->checkint = DEFAULT_CHECKINT;
-	conf->max_checkint = MAX_CHECKINT(conf->checkint);
+	conf->max_checkint = 0;
 	conf->pgfailback = DEFAULT_FAILBACK;
 	conf->fast_io_fail = DEFAULT_FAST_IO_FAIL;
 	conf->retain_hwhandler = DEFAULT_RETAIN_HWHANDLER;
 	conf->detect_prio = DEFAULT_DETECT_PRIO;
+	conf->force_sync = DEFAULT_FORCE_SYNC;
+	conf->partition_delim = DEFAULT_PARTITION_DELIM;
+	conf->processed_main_config = 0;
+	conf->find_multipaths = DEFAULT_FIND_MULTIPATHS;
+	conf->uxsock_timeout = DEFAULT_REPLY_TIMEOUT;
+	conf->uid_attribute = set_default(DEFAULT_UID_ATTRIBUTE);
+	conf->retrigger_tries = DEFAULT_RETRIGGER_TRIES;
+	conf->retrigger_delay = DEFAULT_RETRIGGER_DELAY;
+	conf->uev_wait_timeout = DEFAULT_UEV_WAIT_TIMEOUT;
+	conf->deferred_remove = DEFAULT_DEFERRED_REMOVE;
+	conf->skip_kpartx = DEFAULT_SKIP_KPARTX;
+	conf->disable_changed_wwids = DEFAULT_DISABLE_CHANGED_WWIDS;
 
 	/*
 	 * preload default hwtable
@@ -577,13 +636,13 @@ load_config (char * file, struct udev *udev)
 	/*
 	 * read the config file
 	 */
-	set_current_keywords(&conf->keywords);
-	alloc_keywords();
+	conf->keywords = vector_alloc();
+	init_keywords(conf->keywords);
 	if (filepresent(file)) {
 		int builtin_hwtable_size;
 
 		builtin_hwtable_size = VECTOR_SIZE(conf->hwtable);
-		if (init_data(file, init_keywords)) {
+		if (process_file(conf, file)) {
 			condlog(0, "error parsing config file");
 			goto out;
 		}
@@ -595,13 +654,19 @@ load_config (char * file, struct udev *udev)
 			factorize_hwtable(conf->hwtable, builtin_hwtable_size);
 		}
 
-	} else {
-		init_keywords();
 	}
 
+	conf->processed_main_config = 1;
+	if (conf->config_dir == NULL)
+		conf->config_dir = set_default(DEFAULT_CONFIG_DIR);
+	if (conf->config_dir && conf->config_dir[0] != '\0')
+		process_config_dir(conf, conf->keywords, conf->config_dir);
+
 	/*
 	 * fill the voids left in the config file
 	 */
+	if (conf->max_checkint == 0)
+		conf->max_checkint = MAX_CHECKINT(conf->checkint);
 	if (conf->blist_devnode == NULL) {
 		conf->blist_devnode = vector_alloc();
 
@@ -668,9 +733,8 @@ load_config (char * file, struct udev *udev)
 	    !conf->wwids_file)
 		goto out;
 
-	return 0;
+	return conf;
 out:
 	free_config(conf);
-	return 1;
+	return NULL;
 }
-
