@@ -3,6 +3,7 @@
 
 #include <sys/types.h>
 #include <stdint.h>
+#include <urcu.h>
 
 #define ORIGIN_DEFAULT 0
 #define ORIGIN_CONFIG  1
@@ -19,7 +20,20 @@ enum devtypes {
 	DEV_NONE,
 	DEV_DEVT,
 	DEV_DEVNODE,
-	DEV_DEVMAP
+	DEV_DEVMAP,
+	DEV_UEVENT
+};
+
+enum mpath_cmds {
+	CMD_NONE,
+	CMD_CREATE,
+	CMD_DRY_RUN,
+	CMD_LIST_SHORT,
+	CMD_LIST_LONG,
+	CMD_VALID_PATH,
+	CMD_REMOVE_WWID,
+	CMD_RESET_WWIDS,
+	CMD_ADD_WWID,
 };
 
 struct hwentry {
@@ -48,6 +62,10 @@ struct hwentry {
 	int user_friendly_names;
 	int retain_hwhandler;
 	int detect_prio;
+	int deferred_remove;
+	int delay_watch_checks;
+	int delay_wait_checks;
+	int skip_kpartx;
 	char * bl_product;
 };
 
@@ -71,18 +89,20 @@ struct mpentry {
 	int flush_on_last_del;
 	int attribute_flags;
 	int user_friendly_names;
+	int deferred_remove;
+	int delay_watch_checks;
+	int delay_wait_checks;
+	int skip_kpartx;
 	uid_t uid;
 	gid_t gid;
 	mode_t mode;
 };
 
 struct config {
+	struct rcu_head rcu;
 	int verbosity;
-	int dry_run;
-	int list;
 	int pgpolicy_flag;
 	int pgpolicy;
-	enum devtypes dev_type;
 	int minio;
 	int minio_rq;
 	int checkint;
@@ -96,28 +116,37 @@ struct config {
 	int max_fds;
 	int force_reload;
 	int queue_without_daemon;
+	int ignore_wwids;
 	int checker_timeout;
-	int daemon;
-#ifdef USE_SYSTEMD
-	int watchdog;
-#endif
 	int flush_on_last_del;
 	int attribute_flags;
 	int fast_io_fail;
 	unsigned int dev_loss;
 	int log_checker_err;
 	int allow_queueing;
+	int find_multipaths;
 	uid_t uid;
 	gid_t gid;
 	mode_t mode;
-	uint32_t cookie;
 	int reassign_maps;
 	int retain_hwhandler;
 	int detect_prio;
+	int force_sync;
+	int deferred_remove;
+	int processed_main_config;
+	int delay_watch_checks;
+	int delay_wait_checks;
+	int uxsock_timeout;
+	int strict_timing;
+	int retrigger_tries;
+	int retrigger_delay;
+	int ignore_new_devs;
+	int delayed_reconfig;
+	int uev_wait_timeout;
+	int skip_kpartx;
+	int disable_changed_wwids;
 	unsigned int version[3];
 
-	char * dev;
-	struct udev * udev;
 	char * multipath_dir;
 	char * selector;
 	char * uid_attribute;
@@ -130,11 +159,14 @@ struct config {
 	char * prio_args;
 	char * checker_name;
 	char * alias_prefix;
+	char * partition_delim;
+	char * config_dir;
 	unsigned char * reservation_key;
 
 	vector keywords;
 	vector mptable;
 	vector hwtable;
+	struct hwentry *overrides;
 
 	vector blist_devnode;
 	vector blist_wwid;
@@ -146,11 +178,11 @@ struct config {
 	vector elist_property;
 };
 
-struct config * conf;
+extern struct udev * udev;
 
 struct hwentry * find_hwe (vector hwtable, char * vendor, char * product, char *revision);
-struct mpentry * find_mpe (char * wwid);
-char * get_mpe_wwid (char * alias);
+struct mpentry * find_mpe (vector mptable, char * wwid);
+char * get_mpe_wwid (vector mptable, char * alias);
 
 struct hwentry * alloc_hwe (void);
 struct mpentry * alloc_mpe (void);
@@ -162,8 +194,10 @@ void free_mptable (vector mptable);
 
 int store_hwe (vector hwtable, struct hwentry *);
 
-int load_config (char * file, struct udev * udev);
+struct config *load_config (char * file);
 struct config * alloc_config (void);
 void free_config (struct config * conf);
+extern struct config *get_multipath_config(void);
+extern void put_multipath_config(struct config *);
 
 #endif
