@@ -1,14 +1,18 @@
 #ifndef _UTIL_H
 #define _UTIL_H
 
-int strcmp_chomp(char *, char *);
-void strchop(char *);
-void basenamecpy (char * src, char * dst);
+#include <sys/types.h>
+
+size_t strchop(char *);
+int basenamecpy (const char * src, char * dst, int);
 int filepresent (char * run);
 int get_word (char * sentence, char ** word);
 size_t strlcpy(char *dst, const char *src, size_t size);
 size_t strlcat(char *dst, const char *src, size_t size);
-void remove_trailing_chars(char *path, char c);
+int devt2devname (char *, int, char *);
+dev_t parse_devt(const char *dev_t);
+char *convert_dev(char *dev, int is_path_device);
+void setup_thread_attr(pthread_attr_t *attr, size_t stacksize, int detached);
 
 #define safe_sprintf(var, format, args...)	\
 	snprintf(var, sizeof(var), format, ##args) >= sizeof(var)
